/*
 * Mini-ipc: an example freestanding server, based in memory
 *
 * Copyright (C) 2011,2012 CERN (www.cern.ch)
 * Author: Alessandro Rubini <rubini@gnudd.com>
 *
 * This code is copied from trivial-server, and made even more trivial
 */

#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <wrc.h>

#include "minipc.h"

#define RTIPC_EXPORT_STRUCTURES
#include "rt_ipc.h"

#include <softpll_ng.h>


#define CHAN_ALL_BACKUP_FLAGS (CHAN_UPDATE_PHASE | CHAN_BACKUP_UNLOCKED)

static struct rts_pll_state pstate;

static void clear_state()
{
	int i;
	for(i=0;i<RTS_PLL_CHANNELS;i++)
	{
	    pstate.channels[i].priority = -1;
	    pstate.channels[i].phase_setpoint = 0;
	    pstate.channels[i].phase_loopback = 0;
	    pstate.channels[i].flags = CHAN_REF_VALID;
    }
    pstate.flags = 0;
    pstate.port_status = 0;
    pstate.current_ref = REF_NONE;
    pstate.backup_ref = REF_NONE; 
    pstate.old_ref = REF_NONE; 
    pstate.mode = RTS_MODE_DISABLED;
    pstate.ipc_count = 0;
    pstate.switchover_ocured = 0;
    pstate.backup_mask = 0;
}
static void clear_switchover_occured()
{
    pstate.switchover_ocured = 0;
}
static void set_switchover_occured()
{
    //mask with ports to inform about the update
    pstate.switchover_ocured = 1;
}
static void set_current_channel(int channel)
{
    pstate.current_ref = channel;
}
static void set_backup_channel(int channel)
{
    pstate.backup_ref = channel;
}
static void set_old_channel(int channel)
{
    pstate.old_ref = channel;
}
/* Sets the phase setpoint on a given channel */
int rts_adjust_phase(int channel, int32_t phase_setpoint)
{
    static int at_swover_updated =0;
    TRACE("Adjusting phase: channel %d, setpoint=%d ps [switchover occured=%d,"
    " loopback_phase=%d, flags=0x%x, backup update req %d, phase good_value=%d]\n", 
    channel, phase_setpoint, 
    pstate.switchover_ocured,(int)pstate.channels[channel].phase_loopback, 
    pstate.channels[channel].flags, (pstate.channels[channel].flags & CHAN_UPDATE_PHASE) != 0,
    pstate.channels[channel].phase_good_val);

    if(pstate.switchover_ocured == 1 && at_swover_updated == 0)
    {
	  rts_update(); // to make sure all is in sync, i.e. the curr_ref
	  at_swover_updated = 1;
    }
    else if(pstate.switchover_ocured == 0 && at_swover_updated == 1)
		at_swover_updated = 0;
    
    if(pstate.current_ref == channel)
         spll_set_phase_shift(0, phase_setpoint);
    else
	  spll_set_backup_phase_shift(channel, phase_setpoint);

    pstate.channels[channel].phase_setpoint = phase_setpoint;

    /* Notification to higher layers (PPSi) that something is up and backup_state should be
     * read. Sure, we could read backup_state all the time but why to multiply ipc traffic.
     */
    if( (pstate.switchover_ocured == 1 && pstate.current_ref == channel) ||
        (pstate.switchover_ocured == 0 && (pstate.channels[channel].flags & 
         CHAN_ALL_BACKUP_FLAGS)!=0))
    {
	TRACE("check backup state: curr_ref=%d, chan=%d, swovr_occured=%d, update_phase=%d, backup_unlock=%d\n",
	pstate.current_ref, channel, pstate.switchover_ocured, 
	(pstate.channels[channel].flags & CHAN_UPDATE_PHASE) !=0, 
	(pstate.channels[channel].flags & CHAN_BACKUP_UNLOCKED) !=0 );
	return 1;
    }
    return 0;
}

/* Sets the RT subsystem mode (Boundary Clock or Grandmaster) */
int rts_set_mode(int mode)
{
	int i;

	const struct {
		int mode_rt;
		int mode_spll;
		int do_init;
		char *desc;
	} options[] = {
		{ RTS_MODE_GM_EXTERNAL, SPLL_MODE_GRAND_MASTER, 1, "Grand Master (external clock)" },
		{ RTS_MODE_GM_FREERUNNING, SPLL_MODE_FREE_RUNNING_MASTER, 1, "Grand Master (free-running clock)" },
		{ RTS_MODE_BC, SPLL_MODE_SLAVE, 0, "Boundary Clock (slave)" },
		{ RTS_MODE_DISABLED, SPLL_MODE_DISABLED, 1, "PLL disabled" },
		{ 0,0,0, NULL }
	};

	pstate.mode = mode;

	for(i=0;options[i].desc != NULL;i++)
		if(mode == options[i].mode_rt)
		{
			TRACE("RT: Setting mode to %s.\n", options[i].desc);
			if(options[i].do_init)
				spll_init(options[i].mode_spll, 0, 1, -1 /*priority*/);
			else
				spll_init(SPLL_MODE_DISABLED, 0, 0, -1 /*priority*/);
			
			//reset stuff
			set_current_channel(REF_NONE);
			set_backup_channel(REF_NONE); 
			set_old_channel(REF_NONE);
			clear_switchover_occured();
		}

	return 0;
}


/* Manage the backup channel, called by wrsw_hal via IPC */
int rts_backup_channel(int channel, int cmd)
{
	switch(cmd)
	{
		/* activate backup and remember on which port it is*/
		case RTS_BACKUP_CH_LOCK:
		spll_start_backup(channel,0);
		TRACE("RT [backup port]: locked !!! : %d \n", channel);
		break;
		case RTS_BACKUP_CH_ACTIVATE:
		TRACE("switchover detected by wrsw_hall, finally but we are already done |"
		" %d \n", channel);
		set_old_channel(REF_NONE); //ack from HAL !!!! this allows to synchronized between hal and softPLL
		TRACE("RT [backup port]: activated !!! [port %d going down] \n", channel);
		break;
		case RTS_BACKUP_CH_DOWN:
		spll_stop_backup(channel);	
		TRACE("RT [backup port]: down !!! : %d \n", channel);
		break;
	}

	return 0;
}

/* Reference channel configuration (BC mode only) */
int rts_lock_channel(int channel, int priority)
{
	if(pstate.mode != RTS_MODE_BC)
	{
        TRACE("trying to lock while not in slave mode,..\n");
		return -1;
    }


	TRACE("RT [slave]: Locking to: %d (prio %d) current ref %d \n", 
	channel, priority, pstate.current_ref);
	pstate.channels[channel].priority = priority;
	if(priority < 0 || pstate.current_ref == REF_NONE)
	{
		TRACE("RT [slave]: set current_ref [%d] to  %d)\n", pstate.current_ref, channel);
		spll_init(SPLL_MODE_SLAVE, channel, 0, priority);
	}
	else
	{
		spll_start_backup(channel, priority);
		TRACE("RT [backup port]: locked !!! : %d \n", channel);
	}


	return 0;
}

void rts_init(void)
{
    clear_state();
}

void rts_update(void)
{
    uint32_t i;
    int n_ref;
    int enabled;
    uint32_t ret;

    spll_get_num_channels(&n_ref, NULL);
    ret = spll_get_refs(&pstate.current_ref, &pstate.backup_ref, &pstate.backup_mask);
    if(ret != REF_NONE) // switchover occured
    {
		set_switchover_occured(); // this is only for debug printing
		set_old_channel(ret); // this is needed for HAL !!!
		spll_update_backups(pstate.current_ref);//active which used to be backup (backup chan must be stopped)
		TRACE("rts_update 1: switchover detected setting old port=%d | backup_mask=0x%x\n", 
		      ret, pstate.backup_mask);
    }
/**	  
    if(pstate.backup_ref != REF_NONE && spll_check_switchover(pstate.current_ref,pstate.backup_ref) == 1 )
    {
		TRACE_DEV("[rts_update()] \n");
		//TODO: problem when false switchover detection
// 		//sanity check: verify whether the backup (to be current) is physically working
// 		if((0x1 & (spll_get_hw_status()>> pstate.backup_ref)) == 0x0) // fake switchover we are screwed...
// 		{
// 			//the only thing we can do... RESET ...
// 			set_old_channel(REF_NONE);
// 			set_current_channel(REF_NONE);
// 			set_backup_channel(REF_NONE);		  
// 		}
// 		else
		{
			set_old_channel(pstate.current_ref);
			set_current_channel(pstate.backup_ref);
			set_backup_channel(REF_NONE);
			set_switchover_occured();
		}
    }
*/
    pstate.flags = (spll_check_lock(0) ? RTS_DMTD_LOCKED | RTS_REF_LOCKED : 0);


    /* this seems like a hack but... it turns out that there is a considerable time needed
     * to update *_ref after the hw_status_d knows the link is dowwn...  BUG: ??
     * so to make sure the data that goes to wrsw_hal is synched.. we do the trick
     */
/**
    if(pstate.backup_ref  != REF_NONE && pstate.current_ref != REF_NONE) 
       pstate.port_status |= 0x1 << pstate.current_ref;
    if(pstate.backup_ref  == REF_NONE && pstate.current_ref != REF_NONE) 
       pstate.port_status &= ~(0x1 << pstate.backup_ref);
*/
	pstate.port_status = spll_get_hw_status();
	

    for(i=0;i<RTS_PLL_CHANNELS;i++)
    {
#define CH pstate.channels[i]
        CH.flags &= (CHAN_UPDATE_PHASE | CHAN_BACKUP_UNLOCKED); // remember this utnil cleared by rts_adjust_phase()
        CH.phase_loopback = 0;
        CH.phase_current = 0;
//        CH.phase_setpoint = 0;
        CH.phase_loopback = 0;
        CH.phase_good_val = 0;

        if(i >= n_ref)
            CH.flags = CHAN_DISABLED;
        else {
            if(i==pstate.current_ref)
            {
                spll_get_phase_shift(0, &CH.phase_current, NULL,  &CH.phase_good_val);
		            if(spll_shifter_busy(0))
			    {
		            	CH.flags |= CHAN_SHIFTING;
			    }
		}
            else if(0x1 & (pstate.backup_mask >> i)) //if(i==pstate.backup_ref)
            {
                if(spll_get_backup_phase_shift(i, &CH.phase_current, NULL, &CH.phase_good_val))
			CH.flags |= CHAN_UPDATE_PHASE;
		if(spll_check_backup_lock(i))
			CH.flags |= CHAN_BACKUP_UNLOCKED;
		}

            if(spll_read_ptracker(i, &CH.phase_loopback, &enabled))
	            CH.flags |= CHAN_PMEAS_READY;

	          CH.flags |= (enabled ? CHAN_PTRACKER_ENABLED : 0);

        }

#undef CH
    }
//     show_info();
}

void rts_state_info_dump()
{
    TRACE("| HAL[%6d]: act:%2d: bkp: %d, old %d [HW=0x%2x | BKP=0x%x] ", 
    pstate.ipc_count, pstate.current_ref, pstate.backup_ref, pstate.old_ref, 
    pstate.port_status, pstate.backup_mask);
}
/* fixme: this assumes the host is BE */
static int htonl(int i)
{
    return i;
}

int rts_get_backup_state(struct rts_bpll_state *s, int channel)
{
	
	if(channel < 0) return -1;	  
	
	uint32_t active_flags = 0;
	
	if(pstate.switchover_ocured == 1 && pstate.current_ref == channel)
	{
		clear_switchover_occured();
		TRACE("BPLL_SWITCHOVER\n");
		active_flags = BPLL_SWITCHOVER;
		pstate.channels[channel].flags &= ~CHAN_UPDATE_PHASE;
		pstate.channels[channel].flags &= ~CHAN_BACKUP_UNLOCKED;
	}
	if(pstate.channels[channel].flags & CHAN_UPDATE_PHASE)
	{
		TRACE("BPLL_UPDATE_PHASE\n");
		active_flags |= BPLL_UPDATE_PHASE;
		pstate.channels[channel].flags &= ~CHAN_UPDATE_PHASE;
	}
	if(pstate.channels[channel].flags & CHAN_BACKUP_UNLOCKED)
	{
		TRACE("CHAN_BACKUP_UNLOCKED\n");
		active_flags |= BPLL_UNLOCKED;
		pstate.channels[channel].flags &= ~CHAN_BACKUP_UNLOCKED;
	}
	s->flags         = htonl(active_flags);
	s->phase_good_val= htonl(pstate.channels[channel].phase_good_val);
	s->active_chan   = htonl(pstate.current_ref);
	TRACE("Backup state ret: flags=0x%x, phase_good_val=%d, active_chan=%d \n",
	      s->flags, s->phase_good_val, s->active_chan);
	return 0;
}



static int rts_get_state_func(const struct minipc_pd *pd, uint32_t *args, void *ret)
{
    struct rts_pll_state *tmp = (struct rts_pll_state *)ret;
    int i;

		pstate.ipc_count++;

    /* gaaaah, somebody should write a SWIG plugin for generating this stuff. */
    tmp->current_ref = htonl(pstate.current_ref);
    tmp->backup_ref = htonl(pstate.backup_ref);
    tmp->old_ref = htonl(pstate.old_ref);
    tmp->backup_mask = htonl(pstate.backup_mask);
    tmp->flags = htonl(pstate.flags);
    tmp->port_status = htonl(pstate.port_status);
    tmp->holdover_duration = htonl(pstate.holdover_duration);
    tmp->mode = htonl(pstate.mode);
		tmp->delock_count = spll_get_delock_count();
		tmp->ipc_count = pstate.ipc_count;
		
    for(i=0; i<RTS_PLL_CHANNELS;i++)
    {
        tmp->channels[i].priority = htonl(pstate.channels[i].priority);
        tmp->channels[i].phase_setpoint = htonl(pstate.channels[i].phase_setpoint);
        tmp->channels[i].phase_current = htonl(pstate.channels[i].phase_current);
        tmp->channels[i].phase_loopback = htonl(pstate.channels[i].phase_loopback);
        tmp->channels[i].phase_good_val = htonl(pstate.channels[i].phase_good_val);
        tmp->channels[i].flags = htonl(pstate.channels[i].flags);
    }
    return 0;
}

static int rts_get_backup_state_func(const struct minipc_pd *pd, uint32_t *args, void *ret)
{
	struct rts_bpll_state *state = (struct rts_bpll_state *)ret;
	rts_get_backup_state(state, args[0]);
	return 0;
}


static int rts_set_mode_func(const struct minipc_pd *pd, uint32_t *args, void *ret)
{
		pstate.ipc_count++;
    *(int *) ret = rts_set_mode(args[0]);
    return 0;
}


static int rts_lock_channel_func(const struct minipc_pd *pd, uint32_t *args, void *ret)
{
		pstate.ipc_count++;
    *(int *) ret = rts_lock_channel(args[0], (int)args[1]);
    return 0;
}

static int rts_adjust_phase_func(const struct minipc_pd *pd, uint32_t *args, void *ret)
{
		pstate.ipc_count++;
    *(int *) ret = rts_adjust_phase((int)args[0], (int)args[1]);
    return 0;
}

static int rts_enable_ptracker_func(const struct minipc_pd *pd, uint32_t *args, void *ret)
{
		pstate.ipc_count++;
		spll_enable_ptracker((int)args[0], (int)args[1]);
    *(int *) ret = 0;
    return 0;
}

static int rts_debug_command_func(const struct minipc_pd *pd, uint32_t *args, void *ret)
{
		pstate.ipc_count++;
    *(int *) ret = rts_debug_command((int)args[0], (int)args[1]);
    return 0;
}

static int rts_backup_channel_func(const struct minipc_pd *pd, uint32_t *args, void *ret)
{
		pstate.ipc_count++;
    *(int *) ret = rts_backup_channel((int)args[0], (int)args[1]);
    return 0;
}

static struct minipc_ch *server;

int rtipc_init(void)
{
	/* The mailbox is mapped at 0xF000 in the linker script */
	server = minipc_server_create("mem:F000", 0);
	if (!server)
		return 1;

	rtipc_rts_set_mode_struct.f = rts_set_mode_func;
	rtipc_rts_get_state_struct.f = rts_get_state_func;
	rtipc_rts_lock_channel_struct.f = rts_lock_channel_func;
	rtipc_rts_adjust_phase_struct.f = rts_adjust_phase_func;
	rtipc_rts_enable_ptracker_struct.f = rts_enable_ptracker_func;
	rtipc_rts_debug_command_struct.f = rts_debug_command_func;
	rtipc_rts_backup_channel_struct.f = rts_backup_channel_func;
	rtipc_rts_get_backup_state_struct.f = rts_get_backup_state_func;
	
	minipc_export(server, &rtipc_rts_set_mode_struct);
	minipc_export(server, &rtipc_rts_get_state_struct);
	minipc_export(server, &rtipc_rts_lock_channel_struct);
  minipc_export(server, &rtipc_rts_adjust_phase_struct);
  minipc_export(server, &rtipc_rts_enable_ptracker_struct);
  minipc_export(server, &rtipc_rts_debug_command_struct);
  minipc_export(server, &rtipc_rts_backup_channel_struct);
  minipc_export(server, &rtipc_rts_get_backup_state_struct);


	return 0;
}


void rtipc_action(void)
{
		minipc_server_action(server, 1000);
}
