#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <wrc.h>
#include "board.h"
#include "trace.h"
#include "hw/softpll_regs.h"
#include "hw/pps_gen_regs.h"

#include "softpll_ng.h"

#include "irq.h"

volatile int irq_count = 0;

volatile struct SPLL_WB *SPLL;
volatile struct PPSG_WB *PPSG;

int spll_n_chan_ref, spll_n_chan_out;
int spll_current_ref, spll_backup_ref;
// int fifo_cnt = 0;
/*
 * The includes below contain code (not only declarations) to enable
 * the compiler to inline functions where necessary and save some CPU
 * cycles
 */



#define MAIN_CHANNEL (spll_n_chan_ref)

#define SEQ_START_EXT 1
#define SEQ_WAIT_EXT 2
#define SEQ_START_HELPER 3
#define SEQ_WAIT_HELPER 4
#define SEQ_START_MAIN 5
#define SEQ_WAIT_MAIN 6
#define SEQ_DISABLED 7
#define SEQ_READY 8
#define SEQ_CLEAR_DACS 9
#define SEQ_WAIT_CLEAR_DACS 10

#define AUX_DISABLED 1
#define AUX_LOCK_PLL 2
#define AUX_ALIGN_PHASE 3
#define AUX_READY 4

// defined also in rt_ipc.h - make sure these are in sync
#define REF_NONE 255 

#include "spll_debug.h"

static const struct stringlist_entry seq_states [] =
{
	{ SEQ_START_EXT,       "start-ext  " },
	{ SEQ_WAIT_EXT,        "wait-ext   " },
	{ SEQ_START_HELPER,    "startHelper" },
	{ SEQ_WAIT_HELPER,     "wait-helper" },
	{ SEQ_START_MAIN,      "start-main " },
	{ SEQ_WAIT_MAIN,       "wait-main  " },
	{ SEQ_DISABLED,        "disabled   " },
	{ SEQ_READY,           "ready      " },
	{ SEQ_CLEAR_DACS,      "clear-dacs " },
	{ SEQ_WAIT_CLEAR_DACS, "waitCLRdacs" },
	{ 0, NULL }
};

static const struct stringlist_entry softpll_modes [] =
{
	{ 0,                             "undifiend(0)" },
	{ SPLL_MODE_GRAND_MASTER,        "GMaster  (1)" },
	{ SPLL_MODE_FREE_RUNNING_MASTER, "FMaster  (2)" },
	{ SPLL_MODE_SLAVE,               "slave    (3)" },
	{ SPLL_MODE_DISABLED,            "disabled (4)" }
};

#define SWOVER_UNDEFINED     0
#define SWOVER_PRE_HW_DETECT 1
#define SWOVER_HW_DETECT     2
static const struct stringlist_entry swovr_types [] =
{
	{ SWOVER_UNDEFINED,    "undifiend source      (0)" },
	{ SWOVER_PRE_HW_DETECT,"pre-hardware detection(1)" },
	{ SWOVER_HW_DETECT,    "hardware detection    (2)" }};


#define ALIGN_STATE_EXT_OFF 0
#define ALIGN_STATE_START 1
#define ALIGN_STATE_INIT_CSYNC 2
#define ALIGN_STATE_WAIT_CSYNC 3
#define ALIGN_STATE_START_ALIGNMENT 7
#define ALIGN_STATE_WAIT_SAMPLE 4
#define ALIGN_STATE_COMPENSATE_DELAY 5
#define ALIGN_STATE_LOCKED 6
#define ALIGN_STATE_START_MAIN 8

static const struct stringlist_entry align_states [] =
{
	{ ALIGN_STATE_EXT_OFF,         "ext off     (0)" },
	{ ALIGN_STATE_START,           "start       (1)" },
	{ ALIGN_STATE_WAIT_CSYNC,      "wait csync  (2)" },
	{ ALIGN_STATE_WAIT_SAMPLE,     "wait sample (3)" },
	{ ALIGN_STATE_COMPENSATE_DELAY,"compesate d (4)" },
	{ ALIGN_STATE_LOCKED,          "locked      (5)" },
	{ ALIGN_STATE_START_ALIGNMENT, "start align (6)" },
	{ ALIGN_STATE_START_MAIN,      "start main  (7)" }
};


static volatile struct softpll_state softpll;

static volatile int ptracker_mask = 0;
/* fixme: should be done by spll_init() but spll_init is called to
 * switch modes (and we won't like messing around with ptrackers
 * there) */


static inline void start_ptrackers(struct softpll_state *s)
{
	int i;
	for (i = 0; i < spll_n_chan_ref; i++)
		if (ptracker_mask & (1 << i))
				ptracker_start(&s->ptrackers[i]);
}

static inline void update_ptrackers(struct softpll_state *s, int tag_value, int tag_source)
{
	if(tag_source > spll_n_chan_ref)
		return;
		
	ptrackers_update(s->ptrackers, tag_value, tag_source);
}

static inline void sequencing_fsm(struct softpll_state *s, int tag_value, int tag_source)
{
	switch (s->seq_state) {
		/* State "Clear DACs": initial SPLL sequnencer state. Brings both DACs (not the AUXs) to the default values
		   prior to starting the SPLL. */
		case SEQ_CLEAR_DACS:
		{
			/* Helper always starts at the maximum value (to make sure it locks on positive offset */
			SPLL->DAC_HPLL = s->helper.pi.y_max;

			/* Main starts at midscale */
			SPLL->DAC_MAIN = (s->mpll.pi.y_max + s->mpll.pi.y_min) / 2;
			
			/* we need tags from at least one channel, so that the IRQ that calls this function
			   gets called again */
			spll_enable_tagger(MAIN_CHANNEL, 1);

			softpll.dac_timeout = timer_get_tics()
				+ TICS_PER_SECOND / 20;
			softpll.seq_state = SEQ_WAIT_CLEAR_DACS;
			
			break;
		}
		
		/* State "Wait until DACs have been cleared". Makes sure the VCO control inputs have stabilized before starting the PLL. */
		case SEQ_WAIT_CLEAR_DACS:
		{
			if (time_after(timer_get_tics(), softpll.dac_timeout))
			{
				if(s->mode == SPLL_MODE_GRAND_MASTER)
					s->seq_state = SEQ_START_EXT;
				else
					s->seq_state = SEQ_START_HELPER;
			}
			break;
		}

		/* State "Disabled". Entered when the whole PLL is off */
		case SEQ_DISABLED:
			break;


		/* State "Start external reference PLL": starts up BB PLL for locking local reference to 10 MHz input */
		case SEQ_START_EXT:
		{
			spll_enable_tagger(MAIN_CHANNEL, 0);
			external_start(&s->ext);

			s->seq_state = SEQ_WAIT_EXT;
			break;
		}

		/* State "Wait until we are locked to external 10MHz clock" */
		case SEQ_WAIT_EXT:
		{
			if (external_locked(&s->ext)) {
				start_ptrackers(s);
				s->seq_state = SEQ_READY;
			}
			break;
		}

		case SEQ_START_HELPER:
		{
			helper_start(&s->helper);

			s->seq_state = SEQ_WAIT_HELPER;
			break;
		}

		case SEQ_WAIT_HELPER:
		{
			if (s->helper.ld.locked && s->helper.ld.lock_changed)
			{
				if (s->mode == SPLL_MODE_SLAVE)
				{
					s->seq_state = SEQ_START_MAIN;
				} else {
					start_ptrackers(s);
					s->seq_state = SEQ_READY;	
				}
			}
			break;
		}

		case SEQ_START_MAIN:
		{
			mpll_start(&s->mpll);
			s->seq_state = SEQ_WAIT_MAIN;
			break;
		}

		case SEQ_WAIT_MAIN:
		{
			if (s->mpll.ld.locked)
			{
				start_ptrackers(s);
				s->seq_state = SEQ_READY;
			}
			break;
		}

		case SEQ_READY:
		{
			if (s->mode == SPLL_MODE_GRAND_MASTER && !external_locked(&s->ext)) {
				s->delock_count++;
				s->seq_state = SEQ_CLEAR_DACS;
			} else if (!s->helper.ld.locked) {
				s->delock_count++;
				s->seq_state = SEQ_CLEAR_DACS;
			} else if (s->mode == SPLL_MODE_SLAVE && !s->mpll.ld.locked) {
				s->delock_count++;
				s->seq_state = SEQ_CLEAR_DACS;
			}
			break;
		}
	}
}
void spll_init_swover(struct spll_switchover_state *s)
{
	s->occured     = 0;
	s->trigger_src = 0;
	s->ms_avg_long = 0;
	s->ms_avg_short= 0;
	s->bs_avg_long = 0;
	s->bs_avg_short= 0;
	s->ms_max      = 0;
	s->ms_min      = 0;
	s->ms_mtied    = 0;
	s->ms_down_qualifier_cnt= 0;
	s->old_active_chan = -1;
	s->new_active_chan = -1;
#ifdef MULTI_BACKUP
	int i;
	for (i=0;i<BACKUP_PORTS; i++)
	{
		s->xs_avg_long[i] =0;
		s->xs_avg_short[i] =0;
	}
	s->backup_mask = 0;
#endif
  
}


void spll_update_swover(struct spll_switchover_state *s, struct spll_main_state *mpll, struct spll_backup_state *bpll, int type)
{
	
	s->occured     = 1;
	s->trigger_src = type;
	s->ms_avg_long = avg_get((spll_avg_t *)&mpll->avg_err_long, AVG_HIST_RECENT);
	s->ms_avg_short= avg_get((spll_avg_t *)&mpll->avg_err_short, AVG_HIST_RECENT);
	s->bs_avg_long = avg_get((spll_avg_t *)&bpll->avg_err_long, AVG_HIST_RECENT);
	s->bs_avg_short= avg_get((spll_avg_t *)&bpll->avg_err_short, AVG_HIST_RECENT);
	s->ms_max      = mpll->max;
	s->ms_min      = mpll->min;
	s->ms_mtied    = mpll->mtie_d;
	s->ms_down_qualifier_cnt= mpll->down_qulifier;
	s->old_active_chan = mpll->id_ref;
	s->new_active_chan = bpll->id_ref;
}

int spll_update_dump(struct spll_switchover_state *s)
{
	
	if(s->occured == 0) return 0;
	TRACE_DEV("\n\n\n[switchover occured %2d -> %2d] type: %s | ms_avg_l: %4d |"
	" ms_avg_s: %4d | bs_avg_l: %4d | bs_avg_s: %4d | ms_max: %4d | ms_min: %4d  | "
	"ms_mtie: %4d | ms_down_quantiier: %4d | irq_cnt=%4d ", s->old_active_chan, s->new_active_chan,
	stringlist_lookup(swovr_types,s->trigger_src), s->ms_avg_long,s->ms_avg_short,
	s->bs_avg_long, s->bs_avg_short, s->ms_max, s->ms_min, s->ms_mtied, 
	s->ms_down_qualifier_cnt, s->switchover_irq_cnt);
#ifdef MULTI_BACKUP
	if(s->trigger_src == SWOVER_PRE_HW_DETECT) xpll_switchover_dump(s);
#endif
	TRACE_DEV("\n\n\n\n");
	spll_init_swover(s);
	return 0;
}

int spll_ctr_holdover(struct spll_main_state *ms, struct spll_backup_state *bs)
{
	int avg_l;
	int avg_s;
	if(ms->ld.locked == 0 || bs->ld.locked == 0)
	{
		ms->holdover = 0;
		bs->holdover = 0;
		ms->holdover_cnt=0;
// 		TRACE_DEV("Holdover off -> unlocked\n" );
		return 0;
	}
	else
	{
		avg_l = avg_get((spll_avg_t *)&bs->avg_err_long, AVG_HIST_RECENT);
		avg_s = avg_get((spll_avg_t *)&bs->avg_err_short, AVG_HIST_RECENT);
		
		if(ms->holdover == 1 && ms->holdover == 1 && ms->holdover_cnt < 300)
		{
			ms->holdover_cnt++;
		}
		else if(abs(avg_l-avg_s) > 50 && ms->holdover == 0 && ms->holdover == 0)
		{
			ms->holdover = 1;
			bs->holdover = 1;
			ms->holdover_cnt=0;
// 			TRACE_DEV("Holdover on\n" );
		}
		else if(abs(avg_l-avg_s) < 40 && ms->holdover == 1 && ms->holdover == 1)
		{
			ms->holdover = 0;
			bs->holdover = 0;
			ms->holdover_cnt=0;
// 			TRACE_DEV("Holdover off\n" );
		}
	}
	return 0;
  
}


int spll_predown_detect(struct spll_main_state *ms, struct spll_backup_state *bs)
{
	int bs_avg_l,ms_avg_l;
	int bs_avg_s,ms_avg_s;
	
	if(ms->enabled == 0) return 0;
	
	if(bs->enabled == 0) return 0;
	
	if(ms->ld.locked == 0 || bs->ld.locked == 0)
	{
		ms->min    = 0;
		ms->max    = 0;
		ms->mtie_d = 0;
		ms->down_qulifier = 0;
		bs->stabilize_cntdown = 0x1<<13;
		return 0;
	}
	else if(bs->stabilize_cntdown > 0)
		bs->stabilize_cntdown--;
	else
	{
		bs_avg_l = avg_get((spll_avg_t *)&bs->avg_err_long , AVG_HIST_RECENT);
		bs_avg_s = avg_get((spll_avg_t *)&bs->avg_err_short, AVG_HIST_RECENT);
		ms_avg_l = avg_get((spll_avg_t *)&ms->avg_err_long,  AVG_HIST_RECENT);
		ms_avg_s = avg_get((spll_avg_t *)&ms->avg_err_short, AVG_HIST_RECENT);		

		if(ms_avg_s < ms->min)        ms->min    = ms_avg_s;
		if(ms_avg_s > ms->max)        ms->max    = ms_avg_s;
		if(ms->avg_err_long.acc == 0)
		{ 
			ms->mtie_d = abs(ms->max - ms->min);
			if(ms->mtie_d < 15) ms->mtie_d  = 15; // min mtie
		}
		
		if(ms->down_qulifier > 0)
			ms->down_qulifier--;
		else if(ms->mtie_d != 0 && abs(ms_avg_l - ms_avg_s) > ((3*ms->mtie_d) >> 3)) 
		{
			ms->down_qulifier = 1000;
		}
		
		if(abs(bs_avg_l - bs_avg_s) > 50 && ms->down_qulifier > 0)
		{
// 			ms->min    = 0;
// 			ms->max    = 0;
// 			ms->mtie_d = 0;
// 			ms->down_qulifier = 0;
			return 1;
		}
	}
	return 0;
}

#define AVG_LS_THRESHOLD 40

int spll_multi_predown_detect(struct spll_main_state *ms, struct spll_multibackup_state *bs,
                              struct spll_switchover_state *so)
{
	int bs_avg_l;
	int bs_avg_s;
	int bs_avg_ls;
	int i;
	int detect_mask = 0;
	int cnt_bad = 0;
	
	if(ms->enabled == 0 || bs->backup_number < 2 || bs->backup_mask == 0) return 0;

	for (i=0;i < bs->backup_number; i++)
	{
		bs_avg_ls = bpll_avg_check(&bs->bpll[bs->bids[i]], &bs_avg_l, &bs_avg_s);
		if(bs_avg_ls > AVG_LS_THRESHOLD)
			cnt_bad++;
		so->xs_avg_ls[bs->bids[i]] = bs_avg_ls;
       }
       if (cnt_bad == bs->backup_number)
       {
		so->backup_mask = bs->backup_mask;
		return 1;
       }
       else
		return 0;

/**
	for (i=0;i<BACKUP_ENTRIES_NUM;i++)
	{
		if(0x1 & (bs->backup_mask >> i))
		{
			bs_avg_ls = bpll_avg_check(&bs->bpll[i], &bs_avg_l, &bs_avg_s);
			if(bs_avg_ls > -1 && bs_avg_ls > AVG_LS_THRESHOLD)
			{
// 				TRACE_DEV("backup %d  exceeded  avg l-s: %d | ", i, bs_avg_ls);
// 				avg_dump((spll_avg_t *)&bs->bpll[i].avg_err_long,  "m_err_l");
// 				avg_dump((spll_avg_t *)&bs->bpll[i].avg_err_short, "m_err_s");
// 				TRACE_DEV("\n");
				detect_mask |= (0x1 & (bs->backup_mask >> i));
			}
		}
	  
	}
	if(detect_mask == bs->backup_mask)
		return 1;
	else 
		return 0;
*/
}

static inline void update_loops(struct softpll_state *s, int tag_value, int tag_source)
{
	
	helper_update(&s->helper, tag_value, tag_source);


	if(s->helper.ld.locked)
	{
// 		if(s->bpll.enabled == 1 && spll_predown_detect(&s->mpll, &s->bpll) == 1) 
// 		{
// 			spll_update_swover(&s->swover,&s->mpll, &s->bpll,SWOVER_PRE_HW_DETECT);
// 			s->bpll.enabled = 0;
// 			spll_switchover(s);	
// 		}
// 		else 
#ifdef MULTI_BACKUP
		int bid = spll_backup_ref;//xpll_get_first_backup(&s->xpll);
		if(bid  >=0 && mpll_down(&s->mpll,s->hw_status_d) == 1)
		{
			spll_update_swover(&s->swover,&s->mpll, &s->xpll.bpll[bid],SWOVER_HW_DETECT);
			xpll_update_switchover(&s->swover, &s->xpll);
			spll_switchover(s); // curret bpll gets disabled (enabled=0) so it will be ignored
			s->xpll.backup_mask &= ~(0x1 << bid);
			if(s->xpll.backup_number > 0) s->xpll.backup_number--;
// 			xpll_avg_check_restabilize(&s->xpll);
			s->xpll.bpll[s->xpll.bids[0]].stabilize_cntdown = 0x1<<13; //enought to restabilize single
			s->xpll.after_switchover_cnt = 0x1<<13; 
		}
		else if(spll_multi_predown_detect(&s->mpll, &s->xpll, &s->swover))
		{
			spll_update_swover(&s->swover,&s->mpll, &s->xpll.bpll[bid],SWOVER_PRE_HW_DETECT);
			xpll_update_switchover(&s->swover, &s->xpll);
			spll_switchover(s); // curret bpll gets disabled (enabled=0) so it will be ignored
			s->xpll.backup_mask &= ~(0x1 << bid);
			if(s->xpll.backup_number > 0) s->xpll.backup_number--;
// 			xpll_avg_check_restabilize(&s->xpll);
			s->xpll.bpll[s->xpll.bids[0]].stabilize_cntdown = 0x1<<13; //enought to restabilize single
			s->xpll.after_switchover_cnt = 0x1<<13; 
		}
		
		mpll_update(&s->mpll, tag_value, tag_source);
		xpll_update(&s->xpll, tag_value, tag_source);
#else
		if(s->bpll.enabled  == 1 && mpll_down(&s->mpll,s->hw_status_d) == 1)
		{
			spll_update_swover(&s->swover,&s->mpll, &s->bpll,SWOVER_HW_DETECT);
			s->bpll.enabled = 0;
			spll_switchover(s);
		}
		mpll_update(&s->mpll, tag_value, tag_source);
		bpll_update(&s->bpll, tag_value, tag_source);
#endif

		if(s->seq_state == SEQ_READY) {
			if(s->mode == SPLL_MODE_SLAVE) {
				int i;
				for (i = 0; i < spll_n_chan_out - 1; i++)
					mpll_update(&s->aux[i].pll.dmtd, tag_value, tag_source);
			}
			update_ptrackers(s, tag_value, tag_source);
		}
	}
}
// void show_debug(int irq, struct spll_main_state *ms, struct spll_backup_state *bs,
// 		struct spll_ptracker_state *ptrackers)
// {
// 	int i;
// 	if(ms->enabled)
// 	TRACE_DEV("[mpll %d] tag: out=%d, ref=%d | adder: out=%d, "
// 	          "ref=%d | id: out=%d, ref=%d | err=%d | locked=%d lock_cnt=%d | "
// 	          "gphase_vad = %d | hw_stat=%d | fifo_cnt=%d \n",
// 	           irq, ms->tag_out_d, ms->tag_ref_d, ms->adder_out, ms->adder_ref, 
// 	           ms->id_out, ms->id_ref, ms->err_d, ms->ld.locked, ms->ld.lock_cnt, 
// 	           ms->phase_good_val,ms->hw_status_d,ms->fifo); 
// 	if(bs->enabled)
// 	TRACE_DEV("[bpll %d] tag: out=%d, ref=%d | adder: out=%d, "
// 	          "ref=%d | id: out=%d, ref=%d | err=%d | locked=%d lock_cnt=%d | "
// 	          "gphase_vad = %d \n",
// 	           irq, bs->tag_out_d, bs->tag_ref_d, bs->adder_out, bs->adder_ref, 
// 	           bs->id_out, bs->id_ref, bs->err_d, bs->ld.locked, ms->ld.lock_cnt,
// 	           bs->phase_good_val); 
// 	if(bs->enabled)
// 	TRACE_DEV("[bpll->ld %d] locked=%d, lock_cnt=%d | lock_samples=%d, delock_samples=%d, "
//                  "lock_changed=%d", irq, bs->ld.locked, bs->ld.lock_cnt , bs->ld.lock_samples, 
// 	           bs->ld.delock_samples ,bs->ld.lock_changed);
// 	for(i=0;i<18;i++)
// 	{
// 		register struct spll_ptracker_state *st = ptrackers + i;
// 		if(st->enabled) 
// 		TRACE_DEV("[tracker %d] avg_cnt: %d,  n_avg: %d, acc: %d, ready: %d, "
// 		" phase_val: %d\n", i,st->avg_count,st->n_avg,st->acc, st->ready, 
// 		st->phase_val); 
// 	}
// }
void _irq_entry()
{
	struct softpll_state *s = (struct softpll_state *)&softpll;
	int i = 0;
	s->hw_status_d = SPLL->PSR;
	s->mpll.fifo=0;
	/* check if there are more tags in the FIFO */	
	while (!(SPLL->TRR_CSR & SPLL_TRR_CSR_EMPTY)) {
	
		volatile uint32_t trr = SPLL->TRR_R0;
		int tag_source = SPLL_TRR_R0_CHAN_ID_R(trr);
		int tag_value  = SPLL_TRR_R0_VALUE_R(trr);
		i++;
		if(s->switchover  == 1) continue;
		sequencing_fsm(s, tag_value, tag_source);
		update_loops(s, tag_value, tag_source);
		
	}
	s->mpll.fifo = i;
	spll_current_ref = s->mpll.id_ref;
#ifdef MULTI_BACKUP
	// updated onyl here !!!! to synch | this func also updates list of bacups
	spll_backup_ref  = xpll_get_first_backup(&s->xpll); 
#else
	spll_backup_ref  = s->bpll.id_ref;
#endif

	if(s->switchover==1)
	{
	      s->switchover = 0;
	      s->swover.switchover_irq_cnt = s->mpll.fifo;
	}

	irq_count++;

	clear_irq();
}
void spll_init(int mode, int slave_ref_channel, int align_pps, int priority)
{
	static const char *modes[] = { "", "grandmaster", "freemaster", "slave", "disabled" };
	volatile int dummy;
	int i;

	struct softpll_state *s = (struct softpll_state *) &softpll;

	disable_irq();

	SPLL = (volatile struct SPLL_WB *)BASE_SOFTPLL;
	PPSG = (volatile struct PPSG_WB *)BASE_PPS_GEN;

	uint32_t csr = SPLL->CSR;

	spll_n_chan_ref = SPLL_CSR_N_REF_R(csr); // id of reference channels (max 32, val =18)
	spll_n_chan_out = SPLL_CSR_N_OUT_R(csr); // id of output channels    (max 8, val=1)
	
	spll_current_ref = -1;
	spll_backup_ref = -1;
	
	s->mode = mode;
	s->delock_count = 0;
	s->hw_status_d = SPLL->PSR;
	//to know that we are switching ove
	s->switchover = 0;

	SPLL->DAC_HPLL = 0;
	SPLL->DAC_MAIN = 0;

	SPLL->CSR = 0;
	SPLL->OCER = 0;
	SPLL->RCER = 0;
	SPLL->ECCR = 0;
	SPLL->OCCR = 0;
	SPLL->DEGLITCH_THR = 1000;

	PPSG->ESCR = 0;
	PPSG->CR = PPSG_CR_CNT_EN | PPSG_CR_CNT_RST | PPSG_CR_PWIDTH_W(PPS_WIDTH);

	if(mode == SPLL_MODE_DISABLED)
		s->seq_state = SEQ_DISABLED;
	else
		s->seq_state = SEQ_CLEAR_DACS;

	int helper_ref;
	
	if( mode == SPLL_MODE_SLAVE)
		helper_ref = slave_ref_channel; // Slave mode: lock the helper to an uplink port
	else
		helper_ref = spll_n_chan_ref; // Master/GM mode: lock the helper to the local ref clock

	helper_init(&s->helper, helper_ref);
	mpll_init(&s->mpll, slave_ref_channel, spll_n_chan_ref, priority);
#ifdef MULTI_BACKUP
	xpll_init(&s->xpll);
#else
	bpll_init(&s->bpll);
#endif
	spll_init_swover(&s->swover);
	
	for (i = 0; i < spll_n_chan_out - 1; i++) {
		mpll_init(&s->aux[i].pll.dmtd, slave_ref_channel, spll_n_chan_ref + i + 1,  -1 /*prio*/);
		s->aux[i].seq_state = AUX_DISABLED;
	}
	
	if(mode == SPLL_MODE_FREE_RUNNING_MASTER)
		PPSG->ESCR = PPSG_ESCR_PPS_VALID | PPSG_ESCR_TM_VALID;
	
	for (i = 0; i < spll_n_chan_ref; i++)
		ptracker_init(&s->ptrackers[i], i, PTRACKER_AVERAGE_SAMPLES);

	if(mode == SPLL_MODE_GRAND_MASTER) {
		if(SPLL->ECCR & SPLL_ECCR_EXT_SUPPORTED) {
			s->ext.helper = &s->helper;
			s->ext.main = &s->mpll;
			external_init(&s->ext, spll_n_chan_ref + spll_n_chan_out, align_pps);
		} else {
			TRACE_DEV("softpll: attempting to enable GM mode on non-GM hardware.\n");
			return;
		}
	}

	TRACE_DEV
	    ("softpll: mode %s, %d ref channels, %d out channels\n",
	     modes[mode], spll_n_chan_ref, spll_n_chan_out);

	/* Purge tag buffer */
	while (!(SPLL->TRR_CSR & SPLL_TRR_CSR_EMPTY))
		dummy = SPLL->TRR_R0;
	
	SPLL->EIC_IER = 1;
	SPLL->OCER |= 1;
	
	enable_irq();
}

void spll_shutdown()
{
	disable_irq();

	SPLL->OCER = 0;
	SPLL->RCER = 0;
	SPLL->ECCR = 0;
	SPLL->EIC_IDR = 1;
}

void spll_start_channel(int channel)
{
	struct softpll_state *s = (struct softpll_state *) &softpll;

	if (s->seq_state != SEQ_READY || !channel) {
		TRACE_DEV("Can't start channel %d, the PLL is not ready\n",
			  channel);
		return;
	}
	mpll_start(&s->aux[channel - 1].pll.dmtd);
}

void spll_stop_channel(int channel)
{
	struct softpll_state *s = (struct softpll_state *) &softpll;
	
	if (!channel)
		return;

	mpll_stop(&s->aux[channel - 1].pll.dmtd);
}

int spll_ext_locked()
{
	return external_locked( (struct spll_external_state *) &softpll.ext);
}

int spll_check_lock(int channel)
{
	if (!channel)
		return (softpll.seq_state == SEQ_READY);
	else
		return (softpll.seq_state == SEQ_READY)
		    && softpll.aux[channel - 1].pll.dmtd.ld.locked;
}

uint32_t spll_get_hw_status()
{
// 	TRACE_DEV("softpll.hw_status_d=0x%x\n", softpll.hw_status_d);
	return softpll.hw_status_d;
}

static int32_t to_picos(int32_t units)
{
	return (int32_t) (((int64_t) units *
			   (int64_t) CLOCK_PERIOD_PICOSECONDS) >> HPLL_N);
}

/* Channel 0 = local PLL reference, 1...N = aux oscillators */
static void set_phase_shift(int channel, int32_t value_picoseconds)
{
	struct spll_main_state *st = (struct spll_main_state *)
	    (!channel ? &softpll.mpll : &softpll.aux[channel - 1].pll.dmtd);
	mpll_set_phase_shift(st, value_picoseconds);
	softpll.mpll_shift_ps = value_picoseconds;
}

void spll_set_phase_shift(int channel, int32_t value_picoseconds)
{
	int i;
	if (channel == SPLL_ALL_CHANNELS) {
		spll_set_phase_shift(0, value_picoseconds);
		for (i = 0; i < spll_n_chan_out - 1; i++)
			if (softpll.aux[i].seq_state == AUX_READY)
				set_phase_shift(i + 1, value_picoseconds);
	} else
		set_phase_shift(channel, value_picoseconds);
}

static void set_backup_phase_shift(int channel, int32_t value_picoseconds)
{
#ifdef MULTI_BACKUP
	struct spll_multibackup_state *st = (struct spll_multibackup_state *) &softpll.xpll;
	xpll_set_phase_shift(channel, st, value_picoseconds);
#else
	struct spll_backup_state *st = (struct spll_backup_state *) &softpll.bpll;
	bpll_set_phase_shift(st, value_picoseconds);
#endif

}

void spll_set_backup_phase_shift(int channel, int32_t value_picoseconds)
{
	set_backup_phase_shift(channel, value_picoseconds);
}

// int spll_get_role(int channel)
// {
// 	volatile struct softpll_state *s = (struct softpll_state *)&softpll;
// 	if(s->mpll.id_ref == channel)
// 		return CHAN_ACTIVE_REF;
// #ifdef MULTI_BACKUP
// 	if(s->xpll.bpll[channel].enabled)
// #else
// 	if(s->bpll.id_ref == channel)
// #endif
// 		return CHAN_BACKUP_REF;
// 	return 0;
// }

uint32_t spll_get_refs(uint32_t *current, uint32_t *first_backup, uint32_t *backup_mask)
{
	volatile struct softpll_state *s = (struct softpll_state *)&softpll;
	uint32_t old_cur = *current;
	if(s->mpll.enabled)
		*current = (uint32_t)s->mpll.id_ref;
	else
		*current = (uint32_t)REF_NONE;
#ifdef MULTI_BACKUP
	
	if(s->xpll.bpll[s->xpll.bids[0]].enabled)
		*first_backup = (uint32_t)s->xpll.bids[0];
	else
		*first_backup = (uint32_t)REF_NONE;
	*backup_mask = s->xpll.backup_mask;
#else
	if(s->bpll.enabled)
		*first_backup = (uint32_t)s->bpll.id_ref;
	else
		*first_backup = (uint32_t)REF_NONE;
	*backup_mask = (uint32_t)(0x1<<s->bpll.id_ref);
#endif
	if(old_cur !=-1 && old_cur != *current)
		return old_cur;
	else
		return (uint32_t)REF_NONE;
}

// uint32_t spll_get_current_ref()
// {
// 	volatile struct softpll_state *s = (struct softpll_state *)&softpll;
// 	if(s->mpll.enabled)
// 		return (uint32_t)s->mpll.id_ref;
// 	else
// 		return (uint32_t)REF_NONE;
// }
// uint32_t spll_get_backup_refs()
// {
// 	volatile struct softpll_state *s = (struct softpll_state *)&softpll;
// #ifdef MULTI_BACKUP
// 	return s->xpll.backup_mask;
// #else
// 	return (uint32_t)(0x1<<s->bpll.id_ref);
// #endif
// }

void spll_get_phase_shift(int channel, int32_t *current, int32_t *target, int32_t *good_phase_val)
{
	volatile struct spll_main_state *st = (struct spll_main_state *)
	    (!channel ? &softpll.mpll : &softpll.aux[channel - 1].pll.dmtd);
	int div = (DIVIDE_DMTD_CLOCKS_BY_2 ? 2 : 1);
	if (current)
		*current = to_picos(st->phase_shift_current * div);
	if (target)
		*target = to_picos(st->phase_shift_target * div);
	if (good_phase_val)
		*good_phase_val = st->phase_good_val; // in [ps]
}

/*
 * 
 */
int spll_get_backup_phase_shift(int channel, int32_t *current, int32_t *target, int32_t *good_phase_val)
{
#ifdef MULTI_BACKUP
	volatile struct spll_backup_state *st = (struct spll_backup_state *)&softpll.xpll.bpll[channel];
#else
	volatile struct spll_backup_state *st = (struct spll_backup_state *)&softpll.bpll;
#endif
	int div = (DIVIDE_DMTD_CLOCKS_BY_2 ? 2 : 1);
	if (current)
		*current = to_picos(st->phase_shift_current * div);
	if (target)
		*target = to_picos(st->phase_shift_target * div);
	if (good_phase_val)
		*good_phase_val = st->phase_good_val; // in [ps]	
		
	if(st->xpll->phase_update_mask & (0x1 << channel))
	{
		st->xpll->phase_update_mask = st->xpll->phase_update_mask & ~(0x1 << channel);
		return 1;
	}
	else 
		return 0;
}

int spll_read_ptracker(int channel, int32_t *phase_ps, int *enabled)
{
	volatile struct spll_ptracker_state *st = &softpll.ptrackers[channel];
	struct softpll_state *s = (struct softpll_state *)&softpll;
	
	int phase = st->phase_val;
	
// 	/*ML: if the timestamp is not corrected with the phase measurement, PPSi will try
// 	 *    to compensate by setting setpoint. THe phase measurement should be more or 
// 	 *    less the value of the setpoint. 
// 	 *    By making the phase measurement equal to setpoint, there exist a feedback
// 	 *    system: if the setpoint is wrong because of wrong timestamps,, PPSi will try to 
// 	 *    compensate by adjusting the setpoint...
// 	 */
// 	if(s->bpll.enabled && channel == s->bpll.id_ref)
// 	{
// 	  phase = s->bpll.phase_shift_current;
// 	  TRACE_DEV("[b-ptracker] set phase to setpoint: %d \n",phase);
// 	}
	
	if (phase < 0)
		phase += (1 << HPLL_N);
	else if (phase >= (1 << HPLL_N))
		phase -= (1 << HPLL_N);

	if (DIVIDE_DMTD_CLOCKS_BY_2) {
		phase <<= 1;
		phase &= (1 << HPLL_N) - 1;
	}

	*phase_ps = to_picos(phase);

	if (enabled)
		*enabled = ptracker_mask & (1 << st->id) ? 1 : 0;
	return st->ready;
}

void spll_get_num_channels(int *n_ref, int *n_out)
{
	if (n_ref)
		*n_ref = spll_n_chan_ref;
	if (n_out)
		*n_out = spll_n_chan_out;
}

extern void rts_state_info_dump();

int spll_dump_mpll()
{
#define HANG_CNT 1
	static int err_d, hanged;


  	if (softpll.mode > 0 && softpll.mpll.enabled)
	{
	
		if(softpll.mpll.err_d != 0 && err_d == softpll.mpll.err_d && hanged == 0)
		{
			hanged = HANG_CNT;
			//show the stuff, no return
		}
		else
		{
			if(err_d == softpll.mpll.err_d && hanged > 0)
				hanged--;
			else
			{
				err_d  = softpll.mpll.err_d; 
				hanged = HANG_CNT;
			}
			return 0;
		}
	}
	else 	    
	{
		hanged = HANG_CNT;
		return 0;
	}
			
	  TRACE_DEV("\n\n================== serious dump ==================\n"
	  	"1:  %6d, "               // "state = %d, "
	  	"2:  %6d |  %6d |  %6d |, " // "pi: integrator = %d | x = %d | y = %d |, "
	  	   //"ld: lock_cnt = %d | 
	  	   //"lock_samples = %d | 
	  	   //"delock_samples = %d | 
	  	   //"threshold = %d | 
	  	   //"locked= %d | 
	  	   //"lock_changed = %d |, "
	  	"3:  %6d |  %6d |  %6d |  %6d |  %6d |  %6d |, "
	  	"4:  %6d, " //"avg_err_short = %d, " //11
	  	"5:  %6d, " //"avg_err_long = %d, "12
	  	"6:  %6d, " //"avg_y_long = %d, "13
	  	"7:  %6d,\n", //"adder_ref = %d, "14
	  	softpll.mpll.state,
		softpll.mpll.pi.integrator,
		softpll.mpll.pi.x,
		softpll.mpll.pi.y,
		softpll.mpll.ld.lock_cnt,
		softpll.mpll.ld.lock_samples,
		softpll.mpll.ld.delock_samples,
		softpll.mpll.ld.threshold,
		softpll.mpll.ld.locked,
		softpll.mpll.ld.lock_changed,
		softpll.mpll.avg_err_short.prev_avg[0],
		softpll.mpll.avg_err_long.prev_avg[0],
		softpll.mpll.avg_y_long.prev_avg[0],
		softpll.mpll.adder_ref);	     
	  TRACE_DEV(  
	  	"8:  %6d, "  //"adder_out = %d, "15
	  	"9:  %6d, "  //"tag_ref = %d, " 16
	  	"10: %6d, "  //"tag_out = %d, " 17   
	  	"11: %6d, "  //"tag_ref_d = %d, " 18
	  	"12: %6d, " //"tag_out_d = %d, "19
	  	"13: %6d, "  //"seq_ref = %d, " 20
	  	"14: %6d, " //"seq_out = %d, "21
	  	"15: %6d, " //"match_state = %d, "22
	  	"16: %6d, " //"match_seq = %d, "23
	  	"17: %6d, " //"phase_shift_target = %d, "24   
	  	"18: %6d, " //"phase_shift_current = %d, "25  
	  	"19: %6d, " //"after_switchover = %d, "26
	  	"20: %6d, " //"phase_good_val = %d, "27
	  	"21: %6d,\n ", //"id_ref = %d, "28
		softpll.mpll.adder_out, 
		softpll.mpll.tag_ref, 
		softpll.mpll.tag_out, 
		softpll.mpll.tag_ref_d, 
		softpll.mpll.tag_out_d,
		softpll.mpll.seq_ref, 
		softpll.mpll.seq_out,
		softpll.mpll.match_state,
		softpll.mpll.match_seq,
		softpll.mpll.phase_shift_target,
		softpll.mpll.phase_shift_current,
		softpll.mpll.after_switchover,
		softpll.mpll.phase_good_val,
		softpll.mpll.id_ref);
	  
	   TRACE_DEV(
	  	"22: %6d, " //"id_out = %d, "29
	  	"23: %6d, " //"sample_n = %d, "30             
	  	"24: %6d, " //"delock_count = %d, "31
	  	"25: %6d, " //"dac_index = %d, "32
	  	"26: %6d, " //"enabled = %d, "33
	  	"27: %6d, " //"err_d = %d, "34
	  	"28: %6d, " //"err_slow_drift_corr = %d, "35  
	  	"29: %6d, " //"hw_status_d = %d, "36
	  	"30: %6d, " //"fifo = %d, "37
	  	"31: %6d, " //"holdover = %d, "38
	  	"32: %6d, " //"holdover_cnt = %d \n"39
	  	"33: %6d, " //"min = %d \n"40
	  	"34: %6d, " //"max = %d \n"41
	  	"35: %6d, " //"mtie_d = %d \n"42
	  	"36: %6d, " //"down_qulifier = %d \n"43
	  	"37: %6d\n ", //"priority = %d \n",44
		softpll.mpll.id_out,
		softpll.mpll.sample_n,
		softpll.mpll.delock_count,
		softpll.mpll.dac_index,
		softpll.mpll.enabled,
		softpll.mpll.err_d,
		softpll.mpll.err_slow_drift_corr,
		softpll.mpll.hw_status_d,
		softpll.mpll.fifo,
		softpll.mpll.holdover,
		softpll.mpll.holdover_cnt,
		softpll.mpll.min,
		softpll.mpll.max,
		softpll.mpll.mtie_d,
		softpll.mpll.down_qulifier,
		softpll.mpll.priority);
	return 1;
}
void spll_show_stats()
{
	
	if (softpll.mode > 0)
	{
// 		if(spll_dump_mpll()) return;
		TRACE_DEV("softpll[%8d]; seq %s; mode %s; "
		     "algn_state %s; hL-%s; mL-%s; "
		     "hPiY=%d; mPiY=%d; DelCnt=%d; mErr:%6d; ",
// 		     " holdover:%d", fifordcnt:%d, fifords-over_cnt:%d \n",
		     irq_count, 
		     stringlist_lookup(seq_states, softpll.seq_state), 
		     stringlist_lookup(softpll_modes, softpll.mode),
		     stringlist_lookup(align_states, softpll.ext.align_state), 
		     (softpll.helper.ld.locked ? "yes":" no"), 
		     (softpll.mpll.ld.locked ? "yes":" no"),
// 		     (softpll.bpll.ld.locked ? "yes":" no"),
		     softpll.helper.pi.y, softpll.mpll.pi.y,
		     softpll.delock_count,
		     softpll.mpll.err_d
// 		     softpll.bpll.err_d,
// 		     softpll.mpll.holdover,
// 		     softpll.mpll.fifo,
// 		     softpll.switchover_irq_cnt
		    );
// 		    avg_dump((spll_avg_t *)&softpll.mpll.avg_y_long,   "m_y");
// 		    avg_dump((spll_avg_t *)&softpll.helper.avg_y_long,   "h_y");
// 
// 		    avg_dump((spll_avg_t *)&softpll.mpll.avg_err_long, "m_err_l");
// 		    avg_dump((spll_avg_t *)&softpll.mpll.avg_err_short,"m_err_s");

// 		    avg_dump((spll_avg_t *)&softpll.bpll.avg_err_long, "b_err_l");
// 		    avg_dump((spll_avg_t *)&softpll.bpll.avg_err_short,"b_err_s");

// 		    TRACE_DEV("| m_err_max: %4d m_err_min: %4d m_err_mtie: %d \n", 
// 		    softpll.mpll.max,softpll.mpll.min, softpll.mpll.mtie_d);

		rts_state_info_dump();
		
#ifdef MULTI_BACKUP
		xpll_show_stats((struct spll_multibackup_state *)&softpll.xpll);		
#else
		if(softpll.bpll.enabled)
			 bpll_show_stats((struct spll_backup_state *)&softpll.bpll);
#endif
		
		TRACE_DEV("\n");
		spll_update_dump((struct spll_switchover_state *)&softpll.swover);
		
	}
}



int spll_shifter_busy(int channel)
{
	if (!channel)
		return mpll_shifter_busy((struct spll_main_state *)&softpll.mpll);
	else
		return mpll_shifter_busy((struct spll_main_state *)&softpll.aux[channel - 1].pll.dmtd);
}

void spll_enable_ptracker(int ref_channel, int enable)
{
	if (enable) {
		spll_enable_tagger(ref_channel, 1);
		ptracker_start((struct spll_ptracker_state *)&softpll.
			       ptrackers[ref_channel]);
		ptracker_mask |= (1 << ref_channel);
		TRACE_DEV("Enabling ptracker channel: %d\n", ref_channel);

	} else {
		ptracker_mask &= ~(1 << ref_channel);
		if (ref_channel != softpll.mpll.id_ref)
			spll_enable_tagger(ref_channel, 0);
		TRACE_DEV("Disabling ptracker tagger: %d\n", ref_channel);
	}
}

int spll_get_delock_count()
{
	return softpll.delock_count;
}

static inline int aux_locking_enabled(int channel)
{
	uint32_t occr_aux_en = SPLL_OCCR_OUT_EN_R(SPLL->OCCR);
	
	return occr_aux_en & (1 << channel);
}

static inline void aux_set_channel_status(int channel, int locked)
{
	if(!locked)
		SPLL->OCCR &= ~(SPLL_OCCR_OUT_LOCK_W((1 << channel)));
	else
		SPLL->OCCR |= (SPLL_OCCR_OUT_LOCK_W((1 << channel)));
}

int spll_update_aux_clocks()
{
	int ch;

	for (ch = 1; ch < spll_n_chan_out; ch++) 
	{
		struct spll_aux_state *s = (struct spll_aux_state *) &softpll.aux[ch - 1];

		if(s->seq_state != AUX_DISABLED && !aux_locking_enabled(ch))
		{
			TRACE_DEV("softpll: disabled aux channel %d\n", ch);
			spll_stop_channel(ch);
			aux_set_channel_status(ch, 0);
			s->seq_state = AUX_DISABLED;
		}

		switch (s->seq_state) {
			case AUX_DISABLED:
				if (softpll.mpll.ld.locked && aux_locking_enabled(ch)) {
					TRACE_DEV("softpll: enabled aux channel %d\n", ch);
					spll_start_channel(ch);
					s->seq_state = AUX_LOCK_PLL;
				}
				break;

			case AUX_LOCK_PLL:
				if (s->pll.dmtd.ld.locked) {
					TRACE_DEV ("softpll: channel %d locked [aligning @ %d ps]\n", ch, softpll.mpll_shift_ps);
					set_phase_shift(ch, softpll.mpll_shift_ps);
					s->seq_state = AUX_ALIGN_PHASE;
				}

				break;

			case AUX_ALIGN_PHASE:
				if (!mpll_shifter_busy(&s->pll.dmtd)) {
					TRACE_DEV("softpll: channel %d phase aligned\n", ch);
					aux_set_channel_status(ch, 1);
					s->seq_state = AUX_READY;
				}
				break;

			case AUX_READY:
				if (!softpll.mpll.ld.locked || !s->pll.dmtd.ld.locked) {
					TRACE_DEV("softpll: aux channel %d or mpll lost lock\n", ch);
					aux_set_channel_status(ch, 0); 
					s->seq_state = AUX_DISABLED;
				}
				break;
			}
	}
	return 0;
}

int spll_get_aux_status(int channel)
{
	int rval = 0;

	if (softpll.aux[channel].seq_state != AUX_DISABLED)
		rval |= SPLL_AUX_ENABLED;

	if (softpll.aux[channel].seq_state == AUX_READY)
		rval |= SPLL_AUX_LOCKED;

	return rval;
}

const char *spll_get_aux_status_string(int channel)
{
	const char *aux_stat[] = {"disabled", "locking", "aligning", "locked"};
	struct spll_aux_state *s = (struct spll_aux_state* )&softpll.aux[channel];

	switch(s->seq_state)
	{
		case AUX_DISABLED: return aux_stat[0];
		case AUX_LOCK_PLL: return aux_stat[1];
		case AUX_ALIGN_PHASE: return aux_stat[2];
		case AUX_READY: return aux_stat[3];
	}
	return "";
}

int spll_get_dac(int index)
{
	if (index < 0)
		return softpll.helper.pi.y;
	else if (index == 0)
		return softpll.mpll.pi.y;
	else if (index > 0)
		return softpll.aux[index - 1].pll.dmtd.pi.y;
	return 0;
}

void spll_set_dac(int index, int value)
{
	if (index < 0) {
		softpll.helper.pi.y = value;
		SPLL->DAC_HPLL = value;
	} else {
		SPLL->DAC_MAIN =
		    SPLL_DAC_MAIN_DAC_SEL_W(index) | (value & 0xffff);

		if (index == 0)
			softpll.mpll.pi.y = value;
		else if (index > 0)
			softpll.aux[index - 1].pll.dmtd.pi.y = value;
	}
}

void spll_update()
{
	switch(softpll.mode) {
		case SPLL_MODE_GRAND_MASTER:
			external_align_fsm(&softpll.ext);
			break;
	}
	spll_update_aux_clocks();
}

static int spll_measure_frequency(int osc)
{
	volatile uint32_t *reg;

	switch(osc) {
		case SPLL_OSC_REF:
			reg = &SPLL->F_REF;
			break;
		case SPLL_OSC_DMTD:
			reg = &SPLL->F_DMTD;
			break;
		case SPLL_OSC_EXT:
			reg = &SPLL->F_EXT;
			break;
		default:
			return 0;
	}

    timer_delay_ms(2000);
    return (*reg ) & (0xfffffff);
}

static int calc_apr(int meas_min, int meas_max, int f_center )
{
	// apr_min is in PPM

	int64_t delta_low =  meas_min - f_center;
	int64_t delta_hi = meas_max - f_center;
	int ppm_lo, ppm_hi;

	if(delta_low >= 0)
		return -1;
	if(delta_hi <= 0)
		return -1;

	/* __div64_32 divides 64 by 32; result is in the 64 argument. */
	delta_low = -(int64_t)delta_low * 1000000LL;
	__div64_32(&delta_low, f_center);
	ppm_lo = delta_low;

	delta_hi = delta_hi * 1000000LL;
	__div64_32(&delta_hi, f_center);
	ppm_hi = delta_hi;

	return ppm_lo < ppm_hi ? ppm_lo : ppm_hi;
}

void check_vco_frequencies()
{
	disable_irq();

	int f_min, f_max;
	TRACE_DEV("SoftPLL VCO Frequency/APR test:\n");

	spll_set_dac(-1, 0);
	f_min = spll_measure_frequency(SPLL_OSC_DMTD);
	spll_set_dac(-1, 65535);
	f_max = spll_measure_frequency(SPLL_OSC_DMTD);
	TRACE_DEV("DMTD VCO:  Low=%d Hz Hi=%d Hz, APR = %d ppm.\n", f_min, f_max, calc_apr(f_min, f_max, 62500000));

	spll_set_dac(0, 0);
	f_min = spll_measure_frequency(SPLL_OSC_REF);
	spll_set_dac(0, 65535);
	f_max = spll_measure_frequency(SPLL_OSC_REF);
	TRACE_DEV("REF VCO:   Low=%d Hz Hi=%d Hz, APR = %d ppm.\n", f_min, f_max, calc_apr(f_min, f_max, REF_CLOCK_FREQ_HZ));

	f_min = spll_measure_frequency(SPLL_OSC_EXT);
	TRACE_DEV("EXT clock: Freq=%d Hz\n", f_min);
}
/*
 * this is called by the wrsw_hal when it detects failure of the active port and 
 * a backup port is available
 * 
 * new_ref indicates the new active port (which is currently running as backup)
 */
// void spll_switchover(int new_ref)

void spll_switchover(struct softpll_state *s)
{
// 	struct softpll_state *s = (struct softpll_state *) &softpll;
#ifdef MULTI_BACKUP
	int new_ref = s->swover.new_active_chan;//xpll_get_first_backup(&s->xpll);
#else
	int new_ref = s->bpll.id_ref;
#endif
	volatile struct spll_ptracker_state *st = &softpll.ptrackers[new_ref];
	
	//during the count down, only helper pll is updated on interrupt, mpll waits until count 
	//down is finished
	spll_debug(DBG_EVENT | DBG_MAIN,   DBG_EVT_SWITCHOVER, 1);
	spll_debug(DBG_EVENT | DBG_BACKUP, DBG_EVT_SWITCHOVER, 1);
	spll_debug(DBG_EVENT | DBG_HELPER, DBG_EVT_SWITCHOVER, 1);
#ifdef MULTI_BACKUP
	spll_debug(DBG_EVENT | DBG_BACKUP |(0x1<<4) , DBG_EVT_SWITCHOVER, 1);
	spll_debug(DBG_EVENT | DBG_BACKUP |(0x2<<4) , DBG_EVT_SWITCHOVER, 1);
	spll_debug(DBG_EVENT | DBG_BACKUP |(0x3<<4) , DBG_EVT_SWITCHOVER, 1);
#endif
	disable_irq();
	mpll_fast_holdover(&s->mpll);
	helper_fast_holdover(&s->helper);
	/*switch over helper reference*/
	helper_switch_reference(&s->helper,new_ref);
	/*switch over main pll*/
#ifdef MULTI_BACKUP
	mpll_switchover(&s->mpll, &s->xpll.bpll[new_ref], st->phase_val);
#else
	mpll_switchover(&s->mpll, &s->bpll, st->phase_val);
#endif
	s->switchover = 1;
	enable_irq();
	
	//TODO: disable old tagger..
	
	spll_debug(DBG_EVENT | DBG_MAIN,   DBG_EVT_SWITCHOVER, 1);
	spll_debug(DBG_EVENT | DBG_BACKUP, DBG_EVT_SWITCHOVER, 1);	
	spll_debug(DBG_EVENT | DBG_HELPER, DBG_EVT_SWITCHOVER, 1);
#ifdef MULTI_BACKUP
// 	spll_debug(DBG_EVENT | DBG_BACKUP |(0x1<<4) , DBG_EVT_SWITCHOVER, 1);
// 	spll_debug(DBG_EVENT | DBG_BACKUP |(0x2<<4) , DBG_EVT_SWITCHOVER, 1);
// 	spll_debug(DBG_EVENT | DBG_BACKUP |(0x3<<4) , DBG_EVT_SWITCHOVER, 1);
#endif
}
/*
 * called by PPSi from proto-ext-whiterabbit/state-wrs-s-lock.c via the following path
 * 1) ppsi/proto-ext-whiterabbit/state-wrs-s-lock.c: wr_s_lock()->locking_enable()
 * 2) ppsi/time-wrs/wrs-time.c: wrs_locking_enable 
 *    -- here the proper priority is passed down
 *    -- the prio is read from ppi->slave_prio that is set statically by configuration 
 *    -- TODO: the BMCA was also modified to set the proper priority but not tested, surely
 *             this will need more thinking
 *       TODO: so far only single priority is "settable" and only one backup port possible,
 *             the priorities were expected to handle more backup ports, with shuffling of
 *             their priorities. this is to be done when all other works
 * 3) ppsi/wrsw_hal/hal_exports.c: halexp_lock_cmd()
 * 4) ppsi/wrsw_hal/hal_portc: hal_port_start_lock()
 * 5) wrpc-sw/ipc/rt_ipc.c: rts_lock_channel()
 * 6) wrpc-sw/ipc/rt_ipc.c: rts_backup_channel()
 * 7) wrpc-sw/ipc/rt_ipc.c: spll_switchover()
 * TODO: this  "path" seems a bit overkill, can it be simplified ?
 */
void spll_start_backup(int new_ref,int priority)
{
	struct softpll_state *s = (struct softpll_state *) &softpll;
#ifdef MULTI_BACKUP
	xpll_start(&s->xpll,new_ref, s->mpll.id_out, priority);
#else
	bpll_start(&s->bpll,new_ref, s->mpll.id_out, priority);
#endif
	
}
/*
 * called by wrsw_hal.c:static int handle_link_down() when the backup port is detected
 * to have gone down/dead
 */
void spll_stop_backup(int new_ref)
{
	struct softpll_state *s = (struct softpll_state *) &softpll;
#ifdef MULTI_BACKUP
	xpll_stop(&s->xpll,new_ref);
#else
	bpll_stop(&s->bpll);
#endif
	
}
/*
 * updates back port(s) after switchover, if multi-backup, the list is updated and the 
 * previously backup and now active is deleted from the backup list.
 * always clears the data-structure of the removed backup (used only after switchover)
 */
void spll_update_backups(int removed_backup_id)
{
	struct softpll_state *s = (struct softpll_state *) &softpll;
#ifdef MULTI_BACKUP
	xpll_clear(&s->xpll,removed_backup_id);
#else
	bpll_clear(&s->bpll);
#endif
	
	
}

// void show_info()
// {
// 	struct softpll_state *s = (struct softpll_state *)&softpll;
// 	show_debug(irq_count,&s->mpll, &s->bpll,s->ptrackers);
// }

int spll_check_switchover(int current_ref, int backup_ref)
{
	if(current_ref == spll_current_ref && backup_ref == spll_backup_ref)
		return 0;
	else 
	{
		TRACE_DEV("[spll_check_switchover()]current_ref = %d | spll_current_ref =%d"
		" | backup_ref =%d | spll_backup_ref = %d\n",current_ref,spll_current_ref, 
		backup_ref, spll_backup_ref);
		return 1; 
	}
}

int spll_check_backup_lock(int channel)
{
#ifdef MULTI_BACKUP
	return ld_unlock_d((spll_lock_det_t *)&softpll.xpll.bpll[channel].ld);
#else
	return ld_unlock_d((spll_lock_det_t *)&softpll.bpll.ld.locked);
#endif
}